import pandas as pd
import numpy as np

from typing import Any, Dict, Optional, List

from xurpas_data_quality.data.descriptions import TableDescription



def get_table_stats(df: pd.DataFrame)->dict:
    """
    Get the overview statistics of the DataFrame.

    Args:
        df: the DataFrame object
    
    Retunr:
        dictionary object containing the table statistics
    """
    num_variables = len(df.columns)
    missing_cells = df.isnull().sum().sum()
    duplicate_rows = df.duplicated().sum()

    dataset_stats = {
        "dataset_stats": {
            'num_variables': num_variables,
            'missing_cells': missing_cells,
            'missing_cells_perc' : (missing_cells/df.count().sum())*100,
            'duplicate_rows': duplicate_rows,
            'duplicate_rows_perc': (duplicate_rows/len(df))*100,
            'total_memory': df.memory_usage().sum(),
            'ave_memory': df.memory_usage().sum()/len(df)
            },
        'variable_types': {
            "numeric": df.select_dtypes(include=['int64', 'float64']).shape[1],
            "categorical": df.select_dtypes(include=['object']).shape[1]
        }
    }

    return dataset_stats

def get_correlations(df):
    corr_data = df.corr(numeric_only=True).round(3)
    return corr_data

def get_variable_info(column: pd.Series|pd.DataFrame):
    col_len = len(column)
    distinct = len(column.unique())
    missing = column.isnull().sum().sum()
    memory = column.memory_usage()
    infinite = column.isin([np.inf,-np.inf]).sum().sum()

    def _get_percentage(divisor,dividend=col_len):
        return (divisor/dividend)*100
    
    col_stats ={
        "distinct": distinct,
        "distinct_perc": _get_percentage(distinct),
        "missing": missing,
        "missing_perc": _get_percentage(missing),
        "memory": memory,
    }

    if column.dtype != 'object':
        zeros = column.value_counts().get(0, 0)
        col_stats.update({
            "infinite": infinite,
            "infinite_perc": _get_percentage(infinite),
            "mean": column.mean(),
            "minimum": column.min(),
            "maximum": column.max(),
            "zeros": zeros,
            "zero_perc": _get_percentage(zeros),
            
        })
    return col_stats

def get_variable_statistics(series: pd.Series) ->List[Dict]:
    if series.dtype=='float64' or series.dtype=='int64' or series.dtype=='int32':
        def get_MAD(df):
            med = df.median()
            med_list = [abs(item-med) for item in df]

            return pd.Series(med_list).median()

        def get_monotonicity(df):
            if df.is_monotonic_increasing:
                return "Is increasing"
            elif df.is_monotonic_decreasing:
                return "Is decreasing"
            else:
                return "Not Monotonic"
            
        minimum = series.min()
        fifth_percentile = series.quantile(0.05)
        q1 = series.quantile(0.25)
        median = series.quantile(0.5)
        q3 = series.quantile(0.75)
        ninety_fifth_percentile = series.quantile(0.95)
        maximum = series.max()
        stat_range = maximum - minimum
        interquartile_range = q3-q1
        standard_dev = series.std()
        mean = series.mean()
        cv = (standard_dev/mean)
        kurtosis= series.kurtosis()
        mad = get_MAD(series)
        skew = series.skew()
        total_sum = series.sum()
        variance = series.var()
        monotonicity = get_monotonicity(series)
        return [{
            'minimum': minimum,
            '5th_percentile': fifth_percentile,
            'Q1': q1,
            'median': median,
            'Q3': q3,
            '95th_percentile': ninety_fifth_percentile,
            'maximum': maximum,
            'range': stat_range,
            'IQR': interquartile_range,
        }, { 'std_dev': standard_dev,
            'mean': mean,
            'CV': cv,
            'kurtosis': kurtosis,
            'MAD': mad,
            'skew': skew,
            'sum': total_sum,
            'variance': variance,
            'monotonicity': monotonicity}]
    elif series.dtype == 'object':
        return [{
            'max_length': series.str.len().max()},
        {
            'min_length': series.str.len().min()
        }]

def get_extreme_values(data: pd.DataFrame) -> pd.DataFrame:
    data = data.dropna()
    minimum = data.sort_values(by='Value', ascending=True).head(10)
    maximum = data.sort_values(by='Value', ascending=False).head(10)

    return minimum, maximum

def get_variable_frequencies(series: pd.Series, num_rows: int = 10) -> pd.DataFrame:
    series_value_counts = series.value_counts()

    # Create a DataFrame from the value counts
    df_value_counts = pd.DataFrame({'Count': series_value_counts.values}, index=series_value_counts.index)

    # Calculate the frequency and add it to the DataFrame
    df_value_counts['Frequency (%)'] = (df_value_counts['Count'] / series_value_counts.sum()) * 100

    # Add a row for "Other values"
    if len(series_value_counts) > num_rows:
        other_values_count = sum(series_value_counts.values[num_rows:])
        other_values_freq = (other_values_count / series_value_counts.sum()) * 100
        other_values_df = pd.DataFrame({'Count': [other_values_count], 'Frequency (%)': [other_values_freq]}, index=['Other values (' + str(len(series_value_counts) - num_rows) + ')'])
        df_variable_frequencies = pd.concat([df_value_counts.head(num_rows), other_values_df])
    else:
        df_variable_frequencies = df_value_counts.head(num_rows)

    # Set the name of the index
    df_variable_frequencies.index.name='Value'
    df_value_counts.index.name='Value'

    min_val, max_val = get_extreme_values(df_value_counts)
    return df_variable_frequencies, min_val, max_val

def describe(df: pd.DataFrame) -> TableDescription:

    df_stats = get_table_stats(df)

    column_information = {}

    for column in df.columns:
        if df[column].dtype =='object':
            df[column] = df[column].astype(str)

        common, min_values, max_values = get_variable_frequencies(df[column])
        column_information.update({
            column: {"overview": get_variable_info(df[column]),
                     "details": get_variable_statistics(df[column]),
                     "common": common,
                     "min": min_values,
                     "max": max_values}
        })

    return TableDescription(df, df_stats['dataset_stats'], df_stats['variable_types'], column_information)

# variables['sched_man_hour']['overview']


import pandas as pd

from dataclasses import dataclass
from visions import VisionsTypeset, Generic, Integer, Categorical, Date, DateTime
from typing import Dict, List, Callable, Optional

from xurpas_data_quality.data.type_descriptions import describe_numeric, describe_categorical, describe_generic, describe_date
from xurpas_data_quality.data.typeset import XFloat

@dataclass
class TableDescription:
    df: pd.DataFrame
    dataset_statistics: Dict
    correlation: pd.DataFrame
    variable_types: Dict
    variables: Optional[Dict] = None

class Describer():
    """Describer of series. Contains mapping per type"""

    def __init__(self, 
                type_mapping: Dict[str, List[Callable]] = {
                     Generic : [
                         describe_generic,
                     ],
                     XFloat : [
                         describe_generic,
                         describe_numeric
                     ],
                     Integer : [
                         describe_generic,
                         describe_numeric
                     ],
                     Categorical : [
                         describe_generic,
                         describe_categorical
                     ],
                     Date: [
                        describe_generic,
                        describe_date
                     ],
                     DateTime: [
                        describe_generic,
                        describe_date
                     ]}):
        self.mapping = type_mapping
    
    def summarize(self, dtype, series, **kwargs)-> dict:
        # give the series and typeset
        # for every applicable type apply the function to the series
        # update the summary dict with the new
        summary = {}
        if dtype in self.mapping:
            for func in self.mapping[dtype]:
                summary.update(func(series,summary))
        else:
            for func in self.mapping[Generic]:
                summary.update(func(series,summary))

        return summary



import pandas as pd

from xurpas_data_quality.data.typeset import XTypeSet

def check_dtypes(dtypes: dict) -> dict:
    dtypes_mapping = {value.__name__: value for value in XTypeSet.types}

    for key, value in dtypes.items():
        if value in dtypes_mapping:
            dtypes[key] = dtypes_mapping[value]

    return dtypes

def check_col_names(dtypes: dict, cols: pd.Index) -> dict:
    for key in dtypes.keys():
        if key not in cols:
            raise ValueError(f'{key} is not a column in the given dataset!')
        
    return dtypes


import numpy as np
import pandas as pd


def compute_histogram(series: pd.Series,
                      name:str="histogram")-> dict:
    summary = {}
    if len(series) == 0:
        return {name: []}
    
    bins = np.histogram_bin_edges(series, bins='auto')
    summary[name] = np.histogram(series, bins=bins)
    return summary


from dataclasses import dataclass
from typing import Any, Dict, Optional

import pandas as pd

@dataclass
class TableDescription:
    df: pd.DataFrame
    df_statistics: Dict
    var_types: Dict
    variables: Dict


from enum import Enum, auto, unique
from typing import Optional, List
from visions import Float, Integer

@unique
class AlertType:
    CONSTANT = auto()

    ZEROS = auto()

    HIGH_CORRELATION = auto()

    HIGH_CARDINALITY = auto()

    IMBALANCE = auto()

    SKEWNESS = auto()

    MISSING_VALUES = auto()

    INFINITE_VALUES = auto()

    DATE = auto()

    UNIFORM = auto()

    UNIQUE = auto()

    CONSTANT_LENGTH = auto()

    REJECTED = auto()
    
    UNSUPPORTED = auto()

    DUPLICATES = auto()

    EMPTY = auto()

class Alert:
    def __init__(
            self,
            alert_type: AlertType,
            values: Optional[dict] = None,
            column_name: Optional[str] = None,
            fields: Optional[set] = None
    ):
        self.alert_type = alert_type
        self.column_name = column_name

    @property
    def alert_name(self) -> str:
        return self.alert_type.name
    
    
    def _get_description(self) -> str:
        return "{} on column: {}".format(self.alert_name, self.column_name)
    
    def __repr__(self):
        return self._get_description()
    
class ConstantAlert(Alert):
    def __init__(self, values: Optional[dict] = None, column_name: Optional[str] = None ):
        super().__init__(
            alert_type= AlertType.CONSTANT,
            values=values,
            column_name=column_name,
            fields={'distinct', 'distinct_perc'},
        )
    
    def _get_description(self) -> str:
        return f"[{self.column_name}] has a constant value"
    
class DuplicatesAlert(Alert):
    def __init__(self, values: Optional[dict] = None, column_name: Optional[str] = None ):
        super().__init__(
            alert_type= AlertType.DUPLICATES,
            values=values,
            column_name=column_name,
            fields={'duplicate_rows', 'duplicate_rows_perc'},
        )
    
    def _get_description(self) -> str:
        return f"[{self.column_name}] has a constant value"
    
class InfiniteAlert(Alert):
    def __init__(self, values: Optional[dict] = None, column_name: Optional[str] = None ):
        super().__init__(
            alert_type= AlertType.INFINITE_VALUES,
            values=values,
            column_name=column_name,
            fields={'infinite', 'infinite_perc'},
        )
    
    def _get_description(self) -> str:
        return f"[{self.column_name}] has a constant value"
    
class MissingAlert(Alert):
    def __init__(self, values: Optional[dict] = None, column_name: Optional[str] = None ):
        super().__init__(
            alert_type= AlertType.MISSING_VALUES,
            values=values,
            column_name=column_name,
            fields={'zeros', 'zeros_perc'},
        )
    
    def _get_description(self) -> str:
        return f"[{self.column_name}] contains missing values"
    
class UniqueAlert(Alert):
    def __init__(self, values: Optional[dict] = None, column_name: Optional[str] = None ):
        super().__init__(
            alert_type= AlertType.UNIQUE,
            values=values,
            column_name=column_name,
            fields={'zeros', 'zeros_perc'},
        )
    
    def _get_description(self) -> str:
        return f"[{self.column_name}] values are all unique"
    
class ZerosAlert(Alert):
    def __init__(self, values: Optional[dict] = None, column_name: Optional[str] = None ):
        super().__init__(
            alert_type= AlertType.ZEROS,
            values=values,
            column_name=column_name,
            fields={'zeros', 'zeros_perc'},
        )

    def _get_description(self) -> str:
        return f"[{self.column_name}] contains zeros"
    

def get_overview_alerts(dataset_stats: dict) -> List[Alert]:
    alerts = []
    
    if dataset_stats['duplicate_rows'] > 0:
        alerts.append(DuplicatesAlert)

    return alerts

def get_general_alerts(series_description: dict) -> List[Alert]:
    alerts = []

    if series_description['distinct'] > 0:
        alerts.append(ConstantAlert)

    if series_description['distinct_perc'] == 100:
        alerts.append(UniqueAlert)

    if series_description['missing'] > 0:
        alerts.append(MissingAlert)
    
    return alerts

def get_numerical_alerts(series_description: dict) -> List[Alert]:
    alerts = []

    if series_description['infinite'] > 0:
        alerts.append(InfiniteAlert)
    if series_description['zeros'] > 0:
        alerts.append(ZerosAlert)

    return alerts

def get_variable_alerts(series_description):
    alerts = []
    
    alerts.extend(get_general_alerts(series_description))

    if series_description['type'] == Float or series_description['type'] == Integer:
        alerts.extend(get_numerical_alerts)

    return alerts

def get_alerts(dataset_stats, series_descriptions):
    alerts = []

    alerts.extend(get_overview_alerts(dataset_stats))

    for col in series_descriptions.keys():
        alerts.extend(get_variable_alerts(series_descriptions[col]))

    return alerts


import pandas as pd
from math import log2, ceil
from typing import List

def cbrt(num)->float:
    if 0<=x: 
        return num**(1./3.)
    else:
        return -(-num)**(1./3.)

def sturges(dataset: pd.Series) -> int:
    return ceil(log2(len(dataset)))

def rice(dataset: pd.Series) -> int:
    return ceil(2*cbrt(len(dataset)))

def freedman_diaconis(dataset: pd.Series) -> int:
    try:
        iqr = dataset.quantile(0.75) - dataset.quantile(0.25)
        n = ceil(2*((iqr)/cbrt(len(dataset))))
        return None if n>100 or n==1 else n
    except:
        return None


import pandas as pd
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable

def get_empty_report(df: pd.DataFrame, name:str)-> HTMLBase:
    """
    Generates an empty report
    """

    content = [
        HTMLContainer(
            type="box",
            name="Sample",
            container_items=[
                HTMLTable(
                    id = "sample",
                    data=df.to_html(classes="table table-sm", border=0)
                )
            ]
        )
    ]
    body = HTMLContainer(type="sections",
                         container_items = content)

    if name is not None:
        return HTMLBase(
            body=body,
            name=name
        )
    
    else:
        return HTMLBase(
            body=body
        )


import warnings
import pandas as pd

from visions import Integer, Float, Date, DateTime

from xurpas_data_quality.data import get_series_descriptions, get_correlations
from xurpas_data_quality.data.describer import TableDescription
from xurpas_data_quality.data import XTypeSet

warnings.filterwarnings('ignore', category=UserWarning)

def get_variable_type_counts(variable_types: dict)-> dict:
    value_counts = {}
    for _, value in variable_types.items():
        if value ==Integer or value==Float:
            value = 'Numerical'
        if value == Date or value==DateTime:
            value = 'Date'
        if value in value_counts:
            value_counts[value] += 1
        else:
            value_counts[value] = 1

    return value_counts
    

def get_overview(df: pd.DataFrame, data_types:dict=None) -> dict:
    """
    Get the overview statistics of the DataFrame.

    Args:
        df: the DataFrame object
    
    Retunr:
        dictionary object containing the table statistics
    """
    num_variables = len(df.columns)
    missing_cells = df.isnull().sum().sum()
    duplicate_rows = df.duplicated().sum()

    if data_types is not None:
        v_types = data_types

        for column in df.columns: # updates the v_types dict with the columns that are not user given data types
            if column not in v_types:
                v_types[column] = XTypeSet.infer_type(df[column])

    else:
        v_types = XTypeSet.infer_type(df)

    dataset_stats = {
        "dataset_statistics": {
            'num_variables': num_variables,
            'missing_cells': missing_cells,
            'missing_cells_perc' : (missing_cells/df.count().sum())*100,
            'duplicate_rows': duplicate_rows,
            'duplicate_rows_perc': (duplicate_rows/len(df))*100,
            'total_memory': df.memory_usage().sum(),
            'ave_memory': df.memory_usage().sum()/len(df)
            },
        'variable_types': get_variable_type_counts(v_types)
    }

    return dataset_stats

def describe(df: pd.DataFrame, data_types:dict=None)-> TableDescription:
    """Gets Description of DataFrame which includes
    an overview, its correlations, and variable descriptions
    
    Args
        df: dataframe to be described

    Returns
        a class containing all the descriptions
    """

    correlation = get_correlations(df)
    if data_types is None:
        overview = get_overview(df)
        variables = get_series_descriptions(df)
    else:
        overview = get_overview(df, data_types)
        variables = get_series_descriptions(df, data_types)

    data = TableDescription(df=df, variables=variables, correlation=correlation, **overview)

    return data


import pandas as pd

from typing import Tuple

def describe_generic(series: pd.Series, summary: dict)-> Tuple[pd.Series, dict]:
    """
    Describes a series with information for any data type.

    Args:
        series: series to describe
        summary: dict containing the descriptions of the series so far

    Return:
        The series and the updated summary dict

    """
    value_counts = series.value_counts()

    series_len = len(series)
    distinct = series.nunique()
    missing = series.isnull().sum().sum()
    memory = series.memory_usage()

    _get_percentage = lambda divisor, dividend=series_len:(divisor/dividend)*100
    
    series_stats ={
        "distinct": distinct,
        "distinct_perc": _get_percentage(distinct),
        "missing": missing,
        "missing_perc": _get_percentage(missing),
        "memory": memory,
        "value_counts": value_counts
    }

    summary.update(series_stats)

    return summary



from abc import ABC, abstractmethod
from typing import Any, Dict,List
from visions import VisionsBaseType

from xurpas_data_quality.render.template_loader import template

class BaseRenderer(ABC):
    def __init__(self, content: Dict[str, Any], name: str = None, id:str=None, type:VisionsBaseType=None,**kwargs):
        self.content = content

        if name is not None:
            self.content['name'] = name
        
        if id is not None:
            self.content['id'] = id

        if type is not None:
            self.type = type

    @property
    def name(self) -> str:
        return self.content.get('name', None)
    
    @property
    def id(self) -> str:
        return self.content.get('id', None)

    @abstractmethod
    def render(self) -> Any:
        pass

class HTMLTable(BaseRenderer):
    def __init__(self, data:Any, id:str=None, name:str=None,**kwargs):
        super().__init__(id=id, name=name,content={'data': data}, **kwargs)

    def render(self):
        return template("table.html").render(**self.content)

        
class HTMLToggle(BaseRenderer):
    def __init__(self, text:str, id:str, **kwargs):
        super().__init__( content={'text': text, 'id': id}, **kwargs)

    def render(self):
        return template("toggle.html").render(**self.content)
        
class HTMLCollapse(BaseRenderer):
    def __init__(self, button: HTMLToggle, body:Any, **kwargs):
        super().__init__({'button':button, 'body':body})
    
    def render(self):
        return template("collapse.html").render(**self.content)
    
class HTMLVariable(BaseRenderer):
    def __init__(self, name, body, bottom=None, **kwargs):
        super().__init__({"variable_name":name, "variable_body": body, "variable_bottom": bottom}, **kwargs)

    def render(self):
        return template("variable.html").render(name = self.content['variable_name'], bottom=self.content['variable_bottom'], content = self.content['variable_body'], type=self.type)
    
class HTMLPlot(BaseRenderer):
    def __init__(self, plot:Any, type="small", **kwargs):
        super().__init__({"plot": plot}, **kwargs)
        if type !="small" and type !="large" and type!='categorical':
            raise ValueError("Plot type should be either 'small' or 'large' or 'categorical'")
        else:
            self.type=type

    def render(self):
        return template("plot.html").render(**self.content, type=self.type)
    

class HTMLDropdown(BaseRenderer):
    def __init__(self, id:str, dropdown_items: List[str], dropdown_content: Any,name:str=None, **kwargs):
        super().__init__(
                        {
                            'dropdown_items': dropdown_items,
                            'dropdown_content': dropdown_content
                        },
                        name,
                        id,
                        **kwargs)
    def render(self):
        return template("dropdown.html").render(**self.content)

class HTMLContainer(BaseRenderer):
    def __init__(self, type:str, container_items:list, name:str=None, id:str=None, col=None, **kwargs):
        super().__init__({'container_items':container_items, 'name': name, 'id':id, 'col':col}, **kwargs)
        self.type = type
    def render(self):
        if self.type == 'box':
            return template("containers/box.html").render(container_items=self.content['container_items'])
        elif self.type == 'column':
            return template("containers/column.html").render(container_items=self.content['container_items'])
        elif self.type == 'sections':
            return template("containers/sections.html").render(
                container_items=self.content['container_items'])
        elif self.type == 'tabs':
            return template("containers/tabs.html").render(**self.content)
        elif self.type == 'default':
            return template("containers/default.html").render(**self.content)
        elif self.type == 'test':
            return template("containers/test.html").render(**self.content)
        else:
            raise ValueError(f"unknown Container ({self.type}) type!")

class HTMLBase(BaseRenderer):
    def __init__(self, body: Any, name: str, **kwargs):
        if name is None:
            name = "Profile Report"
        super().__init__(content={"body":body}, name=name, *kwargs)

    def render(self, **kwargs) -> str:
        nav_items = [self.content['name']]
        return template("base.html").render(**self.content, nav_items=nav_items, **kwargs)


import io
import base64

import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

def create_tiny_histogram(histogram: tuple):
    counts, bins = histogram
    fig, ax = plt.subplots()

    bin_width = bins[1] - bins[0]
    bin_centers = bins[:-1] + bin_width / 2
    ax.bar(bin_centers, counts, width=bin_width)

    ax.set(frame_on=False)
    ax.set_yticks([])

    return fig

def create_horizontal_histogram(value_counts: dict):
    fig, ax = plt.subplots()
    ax.barh(list(value_counts.keys()), list(value_counts.values()))
    ax.set(frame_on=False)
    ax.set_xticks([])

    return fig

def create_histogram(histogram: tuple):
    counts, bins = histogram
    fig, ax = plt.subplots(figsize=(12,4))

    bin_width = bins[1] - bins[0]
    bin_centers = bins[:-1] + bin_width / 2
    ax.bar(bin_centers, counts, width=bin_width)

    return fig

def create_tiny_counts(series):
    # Assuming 'df' is your DataFrame and 'Pclass' is your column

    fig, ax = plt.subplots()
    series.plot(kind='barh', ax=ax)

    return fig

def create_heatmap(df: pd.DataFrame):
    # Create a figure and a set of subplots
    fig, ax = plt.subplots()
    plt.box(False)
    # Create the heatmap
    heatmap = ax.imshow(df, cmap='RdYlBu', interpolation='nearest', vmin=-1.0, vmax=1.0)

    # Set the column names as x-axis labels
    ax.set_xticks(np.arange(len(df.columns)))
    ax.set_xticklabels(df.columns)

    # Set the column names as y-axis labels
    ax.set_yticks(np.arange(len(df.columns)))
    ax.set_yticklabels(df.columns)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

    # Create a colorbar
    cbar = plt.colorbar(heatmap)
    cbar.outline.set_visible(False)

    return fig

def create_distribution_plot(column: pd.Series, subplot_size=5):
    fig, axs = plt.subplots(ncols=2, figsize=(2*subplot_size, subplot_size))

    axs[0].boxplot(column)

    axs[1].violinplot(column)

    fig.suptitle(f'{column.name} distribution')

    return fig

def create_interaction_plot(x: pd.Series, y: pd.Series):
    fig = plt.figure(figsize=(8,4))
    plt.scatter(x,y,edgecolor="white")

    return fig

def create_missing_bar_plot(df: pd.DataFrame):
    non_null_counts = df.count()
    fig, ax = plt.subplots(figsize=(12,4))
    # create a bar chart
    ax.bar(non_null_counts.index, non_null_counts.values)
    ax.set(frame_on=False)
    plt.xticks(rotation=45)

    return fig

def create_distribution_from_dataframe(df: pd.DataFrame):
    if 'employee_code' in df.columns:
        df = df.drop('employee_code', axis=1)

    fig = plt.figure()
    boxplot = df.boxplot(figsize=(15,3))
    return fig

def plot_to_base64(fig: plt.figure, minimal:bool=False):
    buf = io.BytesIO()
    
    if minimal:
        fig.savefig(buf, format="svg", bbox_inches='tight', dpi=80)
    else:    
        fig.savefig(buf, format="svg", bbox_inches='tight')
    
    plt.close(fig)

    data = base64.b64encode(buf.getvalue()).decode("utf8")
    buf.close()
    return "data:image/svg+xml;base64,{}".format(data)


from typing import Callable, Sequence, Set

from visions import create_type
from visions import Categorical, DateTime, Date, Integer, Float, String, Generic
from visions.relations import IdentityRelation, InferenceRelation,TypeRelation
from visions.typesets import StandardSet, VisionsTypeset

import pandas as pd
import warnings
from pandas.api import types as pdt

warnings.filterwarnings("ignore", category=UserWarning, module='visions.typesets.typeset')

"""class Numerical(VisionsBaseType):
    @staticmethod
    def get_relations() -> TypeRelation:
        relations = [
            IdentityRelation(Generic),
            IdentityRelation(Integer),
            IdentityRelation(Float),
        ]
        return relations

    @staticmethod
    def contains_op(series: pd.Series, state: dict) -> bool:
        return pd.api.types.is_integer_dtype(series) or pd.api.types.is_float_dtype(series)"""

def float_is_myfloat(series: pd.Series, state):
    return super().contains_op(series, state) & ~series.str.contains('e', na=False)

def float_to_myfloat(series: pd.Series) -> pd.Series:
    return series.astype(float)

"""XFloat = create_type(
    "Float",
    contains=Float.contains_op,
    identity=[Generic, Float],
    inference=[{
        "relationship": float_is_myfloat,
        "transformer": float_to_myfloat,
        "related_type": Float,
    }]
)"""

XFloat = Float # TODO change XFloat
#types = [XFloat, Integer, String, Categorical, Date, DateTime, Generic]
XTypeSet = StandardSet() + Date

def infer_type(data:pd.DataFrame|pd.Series):
    if isinstance(data, pd.Series):
        if data.dtype == 'object':
            if XTypeSet.infer_type(data) is Float and data.str.contains('e').any():
                return String
            else:
                return XTypeSet.infer_type(data)
        else:
            return XTypeSet.infer_type(data)
        
    elif isinstance(data, pd.DataFrame):
        inferred_types = {}
        for column in data.columns:
            if data[column].dtype == 'object':
                if XTypeSet.infer_type(data[column]) is Float and data[column].str.contains('e').any():
                    inferred_types[column] = String
                else:
                    inferred_types[column] = XTypeSet.infer_type(data[column])
            else:
                inferred_types[column] = XTypeSet.infer_type(data[column])
        
        return inferred_types







from visions import Generic, Integer

from xurpas_data_quality.data.typeset import XFloat
from xurpas_data_quality.render.handler import Handler
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot


@Handler.register(XFloat)
@Handler.register(Integer)
def render_bottom_numerical(data, col_name:str,*args, **kwargs):
    variable_bottom = [
        HTMLContainer(
            type="default",
            name="Statistics",
            id="stats",
            container_items=[
                HTMLContainer(
                    type = "column",
                    container_items = HTMLTable(
                        data = data['quantile_stats'],
                        name = 'Quantile Statistics')
                ),
                HTMLContainer(
                    type = "column",
                    container_items = HTMLTable(
                        data = data['descriptive_stats'],
                        name = 'Descriptive Statistics'
                    )
                )
            ]
        ),
        HTMLPlot(
            name="Histogram",
            type="large",
            id="histo",
            plot=plot_to_base64(create_histogram(data['histogram']))
        )
    ]

    return HTMLContainer(
        type="tabs",
        col=col_name,
        container_items=variable_bottom
    )


from xurpas_data_quality.render.renderer import HTMLBase, HTMLVariable, HTMLTable

def render_generic(data, name):
    table = {
        "Distinct": data['distinct'],
        "Distinct (%)": "{:0.2f}%".format(data['distinct_perc']),
        "Missing": data['missing'],
        "Missing (%)": "{:0.2f}%".format(data["missing_perc"]),
        "Memory size": "{} bytes".format(data['memory'])
    }

    variable_body = {
        'table': HTMLTable(table)
    }
    return HTMLVariable(
        name = name,
        type = data['type'],
        body = variable_body
    )


import pandas as pd

from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot

def render_date(data: dict, name:str)->HTMLBase:
    table_1 = {
        "Distinct": data['distinct'],
        "Distinct (%)": "{:0.2f}%".format(data['distinct_perc']),
        "Missing": data['missing'],
        "Missing (%)": "{:0.2f}%".format(data["missing_perc"]),
        "Memory size": "{} bytes".format(data['memory'])
    }

    table_2 = {
        "Minimum": data['min_date'],
        "Maximum": data['max_date'],
        "Time range": f"{data['time_range']} days"
    }

    variable_body = {
        'table_1': HTMLTable(table_1),
        'table_2': HTMLTable(table_2),
        #'plot': HTMLPlot(plot=plot_to_base64(create_tiny_histogram(data['histogram'])))
    }

    return HTMLVariable(
        name = name,
        type = data['type'],
        body = variable_body
    )


import pandas as pd

from xurpas_data_quality.render.handler import Handler
from xurpas_data_quality.render.render_types.bottom.render_bottom_numerical import render_bottom_numerical
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLCollapse, HTMLToggle
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot

def render_numerical(data: dict, name: str)-> HTMLBase:

    overview_1 = {
        "Distinct": data['distinct'],
        "Distinct (%)": "{:0.2f}%".format(data['distinct_perc']),
        "Missing": data['missing'],
        "Missing (%)": "{:0.2f}%".format(data["missing_perc"]),
        "Infinite": data["infinite"],
        "Infinite (%)":"{:0.2f}%".format(data["infinite_perc"]) ,
        "Mean": "{:0.4f}".format(data["mean"])
    }

    overview_2 = {
        "Minimum": data['minimum'],
        "Maximum": data['maximum'],
        "Zeros": data['zeros'],
        "Zeros (%)": "{:0.2f}%".format(data['zeros_perc']),
        "Negative": data['negative'],
        "Negative (%)": "{:0.2f}%".format(data['negative_perc']),
        "Memory size": "{} bytes".format(data['memory']),
    }

    variable_body = {
        'table_1': HTMLTable(overview_1),
        'table_2': HTMLTable(overview_2),
        'plot': HTMLPlot(plot=plot_to_base64(create_tiny_histogram(data['histogram'])))
    }

    return HTMLVariable(
        name = name,
        type = data['type'],
        body = variable_body,
        bottom=HTMLCollapse(
            HTMLToggle("More details", name),
            render_bottom_numerical(data, name)
        )
    )


from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot
from xurpas_data_quality.render.render import render_report

from dataclasses import fields

def get_detailed_variable_info(df, key:str, variables:dict):
    details = variables['details']
    bottom = [            
        HTMLContainer(
            type="default",
            name="Statistics",
            id= "stats",
            container_items= [
                HTMLContainer(
                    type="column",
                    container_items=
                        HTMLTable(
                            data=details[0],
                            name="Quantile Statistics"
                        )
                    
                ),
                HTMLContainer(
                    type="column",
                    container_items=
                        HTMLTable(
                            data=details[0],
                            name="Descriptive Statistics"
                        )
                )
            ]
        ),
        HTMLPlot(
            name="Histogram",
            type="large",
            id="histo",
            plot=plot_to_base64(create_histogram(df[key]))
        ),
        HTMLTable(
            id = f"{key}-common_values",
            name = "Common Values",
            data= variables["common"].to_html(classes="table table-sm", border=0)),
        HTMLContainer(
            type="tabs",
            name="Extreme Values",
            container_items=[
                HTMLTable(
                    id = f"{key}-min-values",
                    name="Minimum 10 Values",
                    data=variables["min"].to_html(classes="table table-sm", border=0)
                ),
                HTMLTable(
                    id = f"{key}-max-values",
                    name="Maximum 10 Values",
                    data=variables["max"].to_html(classes="table table-sm", border=0)
                )
            ]
        )
    ]
    if df[key].dtype != 'object':
        bottom.append(            
            HTMLPlot(
                name="Distribution",
                id="distribution",
                type="large",
                plot=plot_to_base64(create_distribution_plot(df[key]))
            ))
    return HTMLContainer(
        type="tabs",
        col=key,
        container_items=bottom
    )

def get_variable_data(data: TableDescription) -> list:
    variables = []

    for key, value in data.variables.items():
        split_dict = lambda d: (dict(list(d.items())[:len(d)//2]), dict(list(d.items())[len(d)//2:]))
        table_1, table_2 = split_dict(value['overview'])

        variable_body = {
            'table_1':HTMLTable(table_1),
            'table_2':HTMLTable(table_2),
            'plot': HTMLPlot(plot=plot_to_base64(create_tiny_histogram(data.df[key])))
        }
        btn = HTMLToggle("More details", key)
        variables.append(
                    HTMLVariable(
                        name=key,
                        body = variable_body,
                        bottom = HTMLCollapse(btn, 
                                              get_detailed_variable_info(data.df,
                                                                         key,
                                                                         data.variables[key]))
                    )
                )
    
    return variables

def get_interactions(data:TableDescription):
    df = data.df.select_dtypes(exclude=['object'])
    outer_tabs = []
    for column in df.columns:
        inner_tabs = []
        for inner_col in df.columns:
            inner_tabs.append(
                HTMLContainer(
                    type="default",
                    name= inner_col,
                    id = f"{column}-{inner_col}-interaction-inner",
                    container_items = [HTMLPlot(
                        plot= plot_to_base64(create_interaction_plot(df[inner_col],df[column])),
                        type = "large",
                        name = f"{column}-{inner_col} Interaction Plot",
                        id = f"{column}-{inner_col}_interaction_plot"
                    )]
                )
            )
        outer_tabs.append(
            HTMLContainer(
                type="tabs",
                name = column,
                id = f"{column}-interaction-outer",
                container_items = inner_tabs
            )
        )
    return outer_tabs

def get_report(data: TableDescription,minimal:bool, name:str=None):
    report = render_report(data=data, report_name=name, minimal=minimal)
    return report


def get_report_old(data: TableDescription, name:str=None)-> HTMLBase:
    content = []
    overview_section = HTMLContainer(
        type="box",
        name="Overview",
        container_items = [
            HTMLContainer(
                type="column",
                container_items = HTMLTable(
                    data=data.df_statistics,
                    name="Dataset Statistics"
                )),
            HTMLContainer(
                type="column",
                container_items =  HTMLTable(
                    data=data.var_types,
                    name="Variable Types"
                )
            )
        ]
    )
    dropdown = [HTMLDropdown(
        dropdown_items= list(data.df),
        dropdown_content= HTMLContainer(
            type="default",
            container_items=get_variable_data(data)),
        id="variables-dropdown"
    )]
    variables_section = HTMLContainer(
        type="box",
        name="Variables",
        container_items = dropdown
    )

    corr_df = data.df.corr(numeric_only=True).round(3)
    correlation = HTMLContainer(
        type="box",
        name="Correlation",
        container_items=[
            HTMLContainer(
                type="tabs",
                container_items=[
                    HTMLPlot(plot=plot_to_base64(create_heatmap(corr_df)),
                             type="large",
                             id="corr",
                             name="Heatmap"),
                    HTMLTable(
                        id='sample',
                        name="Table",
                        data=corr_df.to_html(classes="table table-sm", border=0))
                ]
            )
        ]
    )

    samples = HTMLContainer(
        type="box",
        name="Sample",
        container_items=[
            HTMLTable(
                id = "sample",
                data=data.df.head(10).to_html(classes="table table-sm", border=0)
            )
        ]
    )

    interactions = HTMLContainer(
        type="box",
        name="Interactions",
        container_items=[
            HTMLContainer(
                type="tabs",
                container_items= get_interactions(data)
            )
        ]
    )

    """    distribution = HTMLContainer(
            type="box",
            name="Distribution",
            container_items=[
                HTMLPlot(
                    plot= plot_to_base64(create_distribution_from_dataframe(data.df)),
                    type="large"
                )
            ]
        )"""
    
    content.extend([
        overview_section,
        variables_section,
        correlation,
        interactions,
        samples
    ])

    body = HTMLContainer(
        type="sections",
        container_items = content,
    )
    
    if name is not None:
        return HTMLBase(
            body=body,
            name=name
        )
    
    else:
        return HTMLBase(
            body=body
        )


from .describer import Describer
from .typeset import XTypeSet, infer_type
from .series_describe import get_series_descriptions
from .correlations import get_correlations
from .describe_new import describe
from .dataframe import load_dataframe, validate_dataframe
from .utils import check_dtypes, check_col_names


import pandas as pd

from typing import Tuple

from xurpas_data_quality.data import Describer, XTypeSet, infer_type

def describe_series(col_name: str, series: pd.Series, data_types:dict=None)->Tuple[str, dict]:
    describer = Describer()
    if data_types is not None and col_name in data_types:
        series_type = data_types[col_name]

    else:
        series_type = infer_type(series)

    series_description = describer.summarize(dtype=series_type, series=series)
    series_description.update({'type': series_type})

    return col_name, series_description

def get_series_descriptions(df: pd.DataFrame, data_types: dict=None)-> dict:
    """
    Provides a statistical description of each series in a DataFrame

    Args
        df: the dataframe to be described

    Returns
        A dict containing the descriptions of each series, and another dict
        containing the datatypes of each series
    """
    descriptions = {}
    for name, series in df.items():

        col_name, series_description = describe_series(name, series, data_types)
        descriptions.update({col_name: series_description})
        
    return descriptions



import pandas as pd
import numpy as np

from typing import  Dict, Tuple

from xurpas_data_quality.data.algorithms import compute_histogram

def get_variable_statistics(series: pd.Series) ->dict:
    def get_MAD(df):
        med = df.median()
        med_list = [abs(item-med) for item in df]

        return pd.Series(med_list).median()

    def get_monotonicity(df):
        if df.is_monotonic_increasing:
            return "Is increasing"
        elif df.is_monotonic_decreasing:
            return "Is decreasing"
        else:
            return "Not Monotonic"
     
    minimum = series.min()
    fifth_percentile = series.quantile(0.05)
    q1 = series.quantile(0.25)
    median = series.quantile(0.5)
    q3 = series.quantile(0.75)
    ninety_fifth_percentile = series.quantile(0.95)
    maximum = series.max()
    stat_range = maximum - minimum
    interquartile_range = q3-q1
    standard_dev = series.std()
    mean = series.mean()

    if mean == 0:
        cv = "N/A"
    else:
        cv = (standard_dev/mean)

    kurtosis= series.kurtosis()
    mad = get_MAD(series)
    skew = series.skew()
    total_sum = series.sum()
    variance = series.var()
    monotonicity = get_monotonicity(series)
    return {
        'quantile_stats':
            {
                'minimum': minimum,
                '5th_percentile': fifth_percentile,
                'Q1': q1,
                'median': median,
                'Q3': q3,
                '95th_percentile': ninety_fifth_percentile,
                'maximum': maximum,
                'range': stat_range,
                'IQR': interquartile_range,
            }, 
        'descriptive_stats':
            { 
            'std_dev': standard_dev,
            'mean': mean,
            'CV': cv,
            'kurtosis': kurtosis,
            'MAD': mad,
            'skew': skew,
            'sum': total_sum,
            'variance': variance,
            'monotonicity': monotonicity
            }
        }


def describe_numeric(series: pd.Series, summary: Dict) ->  Tuple[pd.Series, Dict]:
    """
    Describes a numeric series.

    Args
        series: a pandas Series with a Numeric dtype to be described
        summary: dict containing the series descriptions so far

    Returns
        The series and the updated summary

    """
    """ensure that series is numeric"""
    series = pd.to_numeric(series, errors='coerce')
    series = series.dropna()


    series_len = len(series)

    _get_percentage = lambda divisor, dividend=series_len:(divisor/dividend)*100

    infinite = series.isin([np.inf,-np.inf]).sum().sum()
    negative = (pd.to_numeric(series, errors='coerce') < 0).sum()


    if 'value_counts' in summary:
        zeros = summary['value_counts'].get(0, 0)
    else:
        zeros = series.value_counts().get(0, 0)

    summary.update({
        "infinite": infinite,
        "infinite_perc": _get_percentage(infinite),
        "mean": series.mean(),
        "minimum": series.min(),
        "maximum": series.max(),
        "zeros": zeros,
        "zeros_perc": _get_percentage(zeros),
        "negative": negative,
        "negative_perc": _get_percentage(negative)
    })

    summary.update(compute_histogram(series.dropna()))
    summary.update(get_variable_statistics(series))

    return summary


import os
import warnings
from pathlib import Path

import pandas as pd

from xurpas_data_quality.report import get_report, get_empty_report
from xurpas_data_quality.data import check_dtypes, describe, load_dataframe,validate_dataframe ,check_col_names

warnings.filterwarnings("ignore", category=UserWarning, module='visions')

class DataReport:
    def __init__(self, file:str=None, 
                 df:pd.DataFrame=None, 
                 report_name:str="Data Report", 
                 file_path:str="report.html",
                 data_types:dict=None,
                 minimal:bool = True):
        """
        Initializes the DataReport object
        Args
            file:        The path of the file you want to analyze. If empty, df parameter must exist.
                         Only supports .csv, .xlsx, .parquet, and .orc file formats
            df:          Pandas DataFrame object of data to be analyzed, If using df, file must be empty.
            report_name: Name of the report. Defaults to 'Data Report'
            file_path:   Path/ directory of where the report is to be saved
            data_types:  A dictionary containing the column names and their corresponding data types.
                         If empty, then the data types will be inferred.
            minimal:     Default is True. Check if you want minimal mode as your data report.
        """

        if file is not None and df is None:
            self.df = df
        elif file is None and df is not None:
            self.df = df
        elif file is not None and df is not None:
            raise KeyError("Only 'file' or 'df' should be used one at a time!")
        elif file is None and df is None:
            raise ValueError("Please provide your data in 'file' or 'df' parameters!")

        self.render_empty = True if not validate_dataframe(df) else False # checks if dataframe is empty
        self.report_name = report_name
        self.minimal = minimal

        if data_types is not None:
            self.data_types = check_dtypes(check_col_names(data_types, df.columns))
        else:
            self.data_types = None

        def has_extension(file_path:str):
            return os.path.splitext(file_path)[1] != '' 
        if has_extension(file_path):
            self.save_path = file_path
        else:
            self.save_path = '/'.join([file_path,"report.html"])
            warnings.warn("File name not provided, saving as {file_path}/report.html")

    def describe_dataframe(self):
        self.description = describe(df=self.df, data_types=self.data_types)

    def get_data_quality_report(self, minimal:bool,report_name=None):
        self.describe_dataframe()
        report = get_report(self.description, minimal=minimal,name=report_name)
        return report.render()
    
    def _render_empty_report(self, name=None):
        report = get_empty_report(self.df, name)
        return report.render()
    
    def to_file(self, report_name:str=None, file_name:str="report.html"):
        output = Path(self.save_path)

        if self.render_empty:
            print(f"saving empty report as {self.save_path}")
            output.write_text(self._render_empty_report(self.report_name), encoding='utf-8')
            print("saved!")

        else:
            print(f"saving as {self.save_path}")
            if self.minimal:
                print("saving minimal version of report!")
                from minify_html import minify
                minified_report = minify(self.get_data_quality_report(self.minimal, self.report_name))
                output.write_text(minified_report, encoding='utf-8')
            else:
                output.write_text(self.get_data_quality_report(self.minimal, self.report_name), encoding='utf-8')
            
            print(f"saved!")


from visions import Date

from xurpas_data_quality.render.handler import Handler
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot

@Handler.register(Date)
def render_bottom_categorical(data, *args, **kwargs):
    return HTMLContainer(type="default")


import pandas as pd
import numpy as np

from typing import  Dict, Tuple

def get_samples(series: pd.Series):
    first_five = series.head(5)

    series_samples = {f"row_{value}": value for value in first_five }

    return series_samples

def get_cat_counts(series:pd.Series):
    counts = series.value_counts()
    percent = series.value_counts(normalize=True)*100
    percent = percent.map("{:.2f}%".format)

    return pd.DataFrame({'Counts': counts, 'Percentages': percent})

def describe_categorical(series: pd.Series, summary: Dict) ->  Tuple[pd.Series, Dict]:
    """
    Describes a categorical series.

    Args
        series: a pandas Series with a Numeric dtype to be described
        summary: dict containing the series descriptions so far

    Returns
        The series and the updated summary

    """
    # make sure that the series has no numbers
    series = series.astype('string')

    max_length = series.str.len().max()
    min_length = series.str.len().min()
    median_length = series.str.len().median()
    mean_length = series.str.len().mean()
    unique = sum(summary['value_counts']==1)
    unique_perc = unique/len(summary['value_counts'])*100
    sample = get_samples(series)
    cat_counts = get_cat_counts(series)

    summary.update({
        "max_length": max_length,
        "min_length": min_length,
        "median_length": median_length,
        "mean_length": mean_length,
        "unique": unique,
        "unique_perc": unique_perc,
        "histogram_counts": series.value_counts().to_dict(),
        "samples": sample,
        "cat_counts": cat_counts
    })


    return summary


import pandas as pd

from xurpas_data_quality.render.renderer import HTMLBase, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse
from xurpas_data_quality.visuals import plot_to_base64, create_horizontal_histogram
from xurpas_data_quality.render.render_types.bottom.render_bottom_categorical import render_bottom_categorical


def render_categorical(data: dict, name:str)->HTMLBase:
    table = {
        "Distinct": data['distinct'],
        "Distinct (%)": "{:0.2f}%".format(data['distinct_perc']),
        "Missing": data['missing'],
        "Missing (%)": "{:0.2f}%".format(data["missing_perc"]),
        "Memory size": "{} bytes".format(data['memory'])
    }
    variable_body = {
        'table': HTMLTable(table),
        'plot': HTMLPlot(plot=plot_to_base64(create_horizontal_histogram(data['histogram_counts'])),
                         type="categorical")
    }

    return HTMLVariable(
        name = name,
        type = data['type'],
        body = variable_body,
        bottom = HTMLCollapse(
            HTMLToggle("More details", name),
            render_bottom_categorical(data,name)
        )
    )


import os
import warnings

import pandas as pd


def load_csv(file_path, **kwargs):
    return pd.read_csv(file_path)

def load_excel(file_path, **kwargs):
    return pd.read_excel(file_path, **kwargs)

def load_parquet(file_path, **kwargs):
    return pd.read_parquet(file_path, **kwargs)

def load_orc(file_path, **kwargs):
    return pd.read_orc(file_path, **kwargs)

def validate_dataframe(df: pd.DataFrame) -> bool:
    """
    Checks if dataframe is empty.

    Args
        df: dataframe to validate

    Returns
        A boolean. False if the dataframe is empty, True if otherwise
    """
    if df.empty:
        warnings.warn("DataFrame is empty! Generating Empty Report")
        return False
    else:
        return True

def load_dataframe(file_path: str) -> pd.DataFrame:
    load_methods = {
    '.csv': load_csv,
    '.xlsx': load_excel,
    '.parquet': load_parquet,
    '.orc': load_orc
    }
    # Get the file extension
    _, file_extension = os.path.splitext(file_path)
    
    # Get the loading method from the dictionary
    load_method = load_methods.get(file_extension)
    
    # If the loading method exists, call it, else raise an error
    if load_method:
        return load_method(file_path)
    else:
        raise ValueError(f'Unsupported file format: {file_extension}')


    



import warnings
from typing import Any

import jinja2
from jinja2 import Environment, PackageLoader, select_autoescape

# initialize jinja environment
env = Environment(loader=PackageLoader("xurpas_data_quality", "render/templates"))

def template(template_name:str)-> jinja2.Template:
    if not template_name.endswith('.html'):
        warnings.warn(f"Template {template_name} does not have the correct file extension '.html'! adding '.html' to template name", stacklevel=2)
        template_name += ".html"

    return env.get_template(template_name)



import pandas as pd

from typing import Any
from visions import Integer, Categorical, Date

from xurpas_data_quality.data.typeset import XFloat
from xurpas_data_quality.data.describer import TableDescription
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64,create_heatmap, create_interaction_plot, create_missing_bar_plot
from xurpas_data_quality.render.render_types import render_numerical, render_categorical, render_date, render_generic


def render_variable(data: dict, column: str):
    if data['type'] == Integer or data['type'] == XFloat:
        return render_numerical(data, column)
    
    elif data['type'] == Categorical:
        return render_categorical(data, column)
    
    elif data['type'] == Date:
        return render_date(data, column)
    
    else:
        return render_generic(data, column)
    
def render_variables_section(data: TableDescription) -> list:
    vars = []
    for key, value in data.variables.items():
        variable = render_variable(value, key)
        vars.append(variable)
    #return [render_variable(value,key) for key, value in data.variables.items()]
    return vars

def render_correlation_section(data: pd.DataFrame, minimal:bool):
    return HTMLContainer(
        type="box",
        name="Correlation",
        container_items=[
            HTMLContainer(
                type="tabs",
                container_items=[
                    HTMLPlot(plot=plot_to_base64(create_heatmap(data), minimal),
                             type="large",
                             id="corr",
                             name="Heatmap"),
                    HTMLTable(
                        id='sample',
                        name="Table",
                        data=data.to_html(classes="table table-sm", border=0))
                ]
            )
        ]
    )

def render_interactions_section(data:pd.DataFrame, minimal:bool)->HTMLBase:
    df = data.select_dtypes(exclude=['object'])
    outer_tabs = []
    for column in df.columns:
        inner_tabs = []
        for inner_col in df.columns:
            inner_tabs.append(
                HTMLContainer(
                    type="default",
                    name= inner_col,
                    id = f"{column}-{inner_col}-interaction-inner",
                    container_items = [HTMLPlot(
                        plot= plot_to_base64(create_interaction_plot(df[inner_col],df[column]), minimal=minimal),
                        type = "large",
                        name = f"{column}-{inner_col} Interaction Plot",
                        id = f"{column}-{inner_col}_interaction_plot"
                    )]
                )
            )
        outer_tabs.append(
            HTMLContainer(
                type="tabs",
                name = column,
                id = f"{column}-interaction-outer",
                container_items = inner_tabs
            )
        )
    return outer_tabs

def render_dropdown_section(items: HTMLBase, names:list,id="variables-dropdown")-> list:
    return [
        HTMLDropdown(
        dropdown_items= names,
        dropdown_content= HTMLContainer(
            type="default",
            container_items= items),
        id=id
        )
    ]

def render_report(data: TableDescription, minimal:bool,report_name: str=None) -> Any:
    content = []
    dataset_statistics = {
        'Number of Variables': data.dataset_statistics['num_variables'],
        'Missing Cells': data.dataset_statistics['missing_cells'],
        'Missing Cells (%)': "{:0.2f}%".format(data.dataset_statistics['missing_cells_perc']),
        'Duplicate Rows': data.dataset_statistics['duplicate_rows'],
        'Duplicate Rows (%)': "{:0.2f}%".format(data.dataset_statistics['duplicate_rows_perc'])
    }

    overview_section = HTMLContainer(
        type="box",
        name="Overview",
        container_items = [
            HTMLContainer(
                type="column",
                container_items = HTMLTable(
                    data= dataset_statistics,
                    name="Dataset Statistics"
                )),
            HTMLContainer(
                type="column",
                container_items =  HTMLTable(
                    data= data.variable_types,
                    name="Variable Types"
                )
            )
        ]
    )

    variables_section = HTMLContainer(
        type = "box",
        name = "Variables",
        container_items = render_dropdown_section(items=render_variables_section(data), names=list(data.df))
    )

    missing_section = HTMLContainer(
        type="box",
        name="Missing",
        container_items=[
            HTMLPlot(plot=plot_to_base64(create_missing_bar_plot(data.df), minimal=minimal),
            type="large",
            id="missingplot",
            name="Missing Bar Plot")]
    )

    correlation_section = render_correlation_section(data.correlation, minimal)
    
    interactions_section = HTMLContainer(
        type="box",
        name="Interactions",
        container_items=[
            HTMLContainer(
                type="tabs",
                container_items= render_interactions_section(data.df, minimal)
            )
        ]
    )
    samples_section = HTMLContainer(
        type="box",
        name="Sample",
        container_items=[
            HTMLTable(
                id = "sample",
                data=data.df.head(10).to_html(classes="table table-sm", border=0)
            )
        ]
    )

    if not minimal:
        content.extend([overview_section, 
                        variables_section,
                        missing_section, 
                        correlation_section, 
                        interactions_section,
                        samples_section])
    
    else:
        content.extend([overview_section, 
                correlation_section,
                missing_section,
                variables_section,
                #interactions_section,
                samples_section])


    body = HTMLContainer(
        type="sections",
        container_items = content
    )

    return HTMLBase(body=body, name='Data Report' if report_name is None else report_name)


from visions import VisionsBaseType

from typing import Any

class Handler:
    handlers = {}
    default_handler = None 

    @classmethod
    def register(cls, keyword):
        def decorator(func):
            cls.handlers[keyword] = func
            return func
        return decorator

    @classmethod
    def register_default(cls):
        def decorator(func):
            cls.default_handler = func
            return func
        return decorator

    @classmethod
    def render_bottom(cls, type, data, *args, **kwargs):
        if type in cls.handlers:
            return cls.handlerstype
        elif cls.default_handler is not None: 
            return cls.default_handler(data)
        else:
            raise ValueError(f"No handler registered for {type}")


from visions import Categorical

from xurpas_data_quality.render.handler import Handler
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot

@Handler.register(Categorical)
def render_bottom_categorical(data, col_name,*args, **kwargs):

    variable_bottom = [
        HTMLContainer(
            type="default",
            name="Overview",
            id = 'overview',
            container_items= [
                HTMLContainer(
                    type = "column",
                    container_items = HTMLTable(
                        data = {
                            "Max Length": data['max_length'],
                            "Median Length": data['median_length'],
                            "Mean Length": data['mean_length'],
                            "Minimum Length": data['min_length']

                        }
                    )
                ),
                HTMLContainer(
                    type ="column",
                    container_items = HTMLTable(
                        data = {f"{i+1}st Row" if i == 0 else f"{i+1}nd Row" if i == 1 else f"{i+1}rd Row" if i == 2 else f"{i+1}th Row": value for i, (key, value) in enumerate(data['samples'].items())}
                    )

                )
            ]
        ),
        HTMLContainer(
            type="default",
            name="Categories",
            id="categories",
            container_items=[
                HTMLTable(
                    id=f"{col_name}-cat-samples",
                    name='Table',
                    data = data['cat_counts'].to_html(classes="table table-sm", border=0)
                )
            ]
        )
    ]
    return HTMLContainer(type="tabs",
                         col=col_name,
                         container_items=variable_bottom)


import os
from setuptools import setup, find_packages

#taken from pip/setup.py
def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

setup(
    name='Xurpas Data Quality Report',
    version=get_version("src/xurpas_data_quality/__init__.py"),
    packages = find_packages(
        where='src',
        include=['xurpas_data_quality.*']),
    package_dir={"":"src"},
    package_data={
        'xurpas_data_quality': ["py.typed"]
    },
    include_package_data=True,
    author='Neil Ortaliz',
    author_email='neillaurenceortaliz@gmail.com',
    description='XAIL Data quality',
    install_requires=[
        'pandas',
        'numpy',
        'matplotlib',
        'jinja2',
        'openpyxl',
        'pyarrow',
        'pytest',
        'minify_html'
    ]
)


import warnings

import pandas as pd

from xurpas_data_quality.data.dataframe import check_df,preprocess

@check_df.register
def check_df_pandas(df: pd.Dataframe) -> None:
    if not isinstance(df, pd.DataFrame):
        warnings.warm("df is not pandas.DataFrame type!")

@preprocess.register
def preprocess_pandas_df() -> None:
    pass


import pandas as pd

from typing import Tuple

def describe_date(series: pd.Series, summary: dict)-> Tuple[pd.Series, dict]:
    """
    Describes a series with information for date data type.

    Args:
        series: series to describe
        summary: dict containing the descriptions of the series so far

    Return:
        The series and the updated summary dict
        """
    series = series.dropna()
    min_date = pd.to_datetime(series.min())
    max_date = pd.to_datetime(series.max())
    med_date = pd.to_datetime(min_date + (max_date-min_date))
    time_range = (max_date - min_date).days
    histogram_counts = series.value_counts().to_dict()
    summary.update(
            {'min_date': min_date,
             'median_date': med_date,
            'max_date': max_date,
            'time_range': time_range,
            'histogram_counts':histogram_counts})
    
    return summary
